/**

Copyright (c) Scott Gasch

Module Name:

    Os-dependant.c

Abstract:

    FreeBSD/Linux/OSX system dependent code.

Author:

    Scott Gasch (SGasch) 8 Apr 2004

Revision History:

    $Id: Os-dependant.c 200 2006-07-05 20:05:07Z scott $

**/

#if defined(__GNUC__) && !defined(__MINGW32__) || defined(__CYGWIN__)

#include "chess.h"
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/select.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <pthread.h>
#include <signal.h>
#include <errno.h>

#define SYS_MAX_HEAP_ALLOC_SIZE_BYTES 0xfff
#define HEAP 0x48656170
#define MMAP 0x4d6d6170

pthread_mutex_t g_SystemLock;
#define LOCK_SYSTEM     pthread_mutex_lock(&g_SystemLock)
#define UNLOCK_SYSTEM   pthread_mutex_unlock(&g_SystemLock)

#ifdef DEBUG
uint32_t g_uTotalAlloced = 0;

#define ALLOC_HASH_SIZE 0x4000

typedef struct _ALLOC_RECORD
{
    void *p;
    uint32_t uSize;
} ALLOC_RECORD;
ALLOC_RECORD g_AllocHash[ALLOC_HASH_SIZE];

#define PTR_TO_ALLOC_HASH(x) ((((uint32_t)(x)) >> 3) & (ALLOC_HASH_SIZE - 1))

uint32_t
GetHeapMemoryUsage(void)
{
    Trace("There are now %u bytes outstanding.\n", g_uTotalAlloced);
    return(g_uTotalAlloced);
}

static void 
ReleaseAllocHashEntry(void *p)
{
    uint32_t u = PTR_TO_ALLOC_HASH(p);
    uint32_t v = u;

    LOCK_SYSTEM;
    while(g_AllocHash[u].p != p)
    {
        u++;
        u &= (ALLOC_HASH_SIZE - 1);
        if (v == u) return;
    }
    g_AllocHash[u].p = NULL;
    g_uTotalAlloced -= g_AllocHash[u].uSize;
    UNLOCK_SYSTEM;

//    Trace("Freed %u bytes at %p.\n", g_AllocHash[u].uSize, p);
}

static void
MarkAllocHashEntry(void *p, uint32_t uSize)
{
    uint32_t u = PTR_TO_ALLOC_HASH(p);

    if (uSize > (1024 * 1024 * 30)) return;

    LOCK_SYSTEM;
    while(g_AllocHash[u].p != NULL)
    {
        u++;
        u &= (ALLOC_HASH_SIZE - 1);
    }
    g_AllocHash[u].p = p;
    g_AllocHash[u].uSize = uSize;
    g_uTotalAlloced += uSize;
    UNLOCK_SYSTEM;

//    Trace("Alloc %u bytes at %p.\n", g_AllocHash[u].uSize, p);
}
#endif // DEBUG

//
// Thread table
//
typedef struct _SYSTEM_THREAD_TABLE_ENTRY
{
    DLIST_ENTRY links;
    uint32_t uWrapperHandle;
    uint32_t uTid;
    uint32_t uThreadParam;
    pthread_t thread;
    THREAD_ENTRYPOINT *pEntry;
    uint32_t uExitCode;
}
SYSTEM_THREAD_TABLE_ENTRY;
DLIST_ENTRY g_SystemThreadList;
uint32_t g_uNextThreadHandle = 6000;

int8_t *
SystemGetTimeString(void)
/**

Routine description:

    Get the current time string from the OS.  Not thread safe.

Parameters:

    void

Return value:

    int8_t

**/
{
    static int8_t buf[32];
    time_t t = time(NULL);
    struct tm *p;

    p = localtime(&t);
    snprintf(buf, ARRAY_LENGTH(buf), 
             "%u:%02u:%02u", p->tm_hour, p->tm_min, p->tm_sec);
    return(buf);
}

int8_t *
SystemGetDateString(void)
/**

Routine description:

    Get the current date string from the OS.  Not thread safe.

Parameters:

    void

Return value:

    int8_t

**/
{
    static int8_t buf[32];
    time_t t = time(NULL);
    struct tm *p;

    p = localtime(&t);
    snprintf(buf, ARRAY_LENGTH(buf), 
             "%u.%02u.%02u", p->tm_year + 1900, p->tm_mon + 1, p->tm_mday);
    return(buf);
}

//
// Thread code
//

static SYSTEM_THREAD_TABLE_ENTRY *
_SystemGetThreadTableEntry(uint32_t uThreadHandle)
/**

Routine description:

    Support routine for thread functionality wrapper.  Given a
    uThreadHandle look up its entry in the global table.

Parameters:

    uint32_t uThreadHandle

Return value:

    SYSTEM_THREAD_TABLE_ENTRY *

**/
{
    DLIST_ENTRY *p;
    SYSTEM_THREAD_TABLE_ENTRY *q;
    
    LOCK_SYSTEM;
    p = g_SystemThreadList.pFlink;
    while(p != &g_SystemThreadList)
    {
        q = CONTAINING_STRUCT(p, SYSTEM_THREAD_TABLE_ENTRY, links);
        if (q->uWrapperHandle == uThreadHandle)
        {
            goto end;
        }
        p = p->pFlink;
    }
    q = NULL;
 end:
    UNLOCK_SYSTEM;
    return(q);
}

void *
SystemThreadEntryPoint(void *pParam)
/**

Routine description:

    The first code that any newly created thread executes.

Parameters:

    void *pParam

Return value:

    void

**/
{
    SYSTEM_THREAD_TABLE_ENTRY *p = (SYSTEM_THREAD_TABLE_ENTRY *)pParam;
    uint32_t uParam;
    int32_t i;
    
    LOCK_SYSTEM;
    UNLOCK_SYSTEM;

    uParam = p->uThreadParam;
    i = (int32_t)(*(p->pEntry))(uParam);          // call thread's user-supplied entry
    return((void *)i);
}

FLAG 
SystemCreateThread(THREAD_ENTRYPOINT *pEntry, uint32_t uParam, uint32_t *puHandle)
/**

Routine description:

    Wraps the OS/Library dependent thread creation call.

Parameters:

    THREAD_ENTRYPOINT *pEntry : where should the new thread start?
    uint32_t uParam : parameter to pass to the new thread
    uint32_t *puHandle : "handle" to the new thread

Return value:

    FLAG

**/
{
    SYSTEM_THREAD_TABLE_ENTRY *p;

    p = malloc(sizeof(SYSTEM_THREAD_TABLE_ENTRY));
    if (NULL == p)
    {
        *puHandle = 0;
        return(FALSE);
    }
    
    LOCK_SYSTEM;
    p->pEntry = pEntry;
    p->uThreadParam = uParam;
    if (0 != pthread_create(&(p->thread),
                            NULL, 
                            SystemThreadEntryPoint, 
                            p))
    {
        free(p);
        return(FALSE);
    }
    *puHandle = p->uWrapperHandle = g_uNextThreadHandle;
    g_uNextThreadHandle++;
    InsertHeadList(&g_SystemThreadList, &(p->links));
    UNLOCK_SYSTEM;
    return(TRUE);
}


FLAG 
SystemWaitForThreadToExit(uint32_t uThreadHandle)
/**

Routine description:

    Blocks until the thread whose handle is provieded exits.

Parameters:

    uint32_t uThreadHandle

Return value:

    FLAG

**/
{
    SYSTEM_THREAD_TABLE_ENTRY *q = _SystemGetThreadTableEntry(uThreadHandle);
    
    pthread_join(q->thread, NULL);
    return(TRUE);
}


FLAG 
SystemGetThreadExitCode(uint32_t uThreadHandle, uint32_t *puCode)
/**

Routine description:

    Get the exit value of an already-exited thread.

Parameters:

    uint32_t uThreadHandle,
    uint32_t *puCode

Return value:

    FLAG

**/
{
    SYSTEM_THREAD_TABLE_ENTRY *q = _SystemGetThreadTableEntry(uThreadHandle);
    void *p;

    pthread_join(q->thread, &p);
    *puCode = ((uint32_t)p);
    return(TRUE);
}


FLAG 
SystemDeleteThread(uint32_t uThreadHandle)
/**

Routine description:

    Kill a thread.

Parameters:

    uint32_t uThreadHandle

Return value:

    FLAG

**/
{
    SYSTEM_THREAD_TABLE_ENTRY *q = _SystemGetThreadTableEntry(uThreadHandle);
    
    if (NULL != q)
    {
        pthread_kill(q->thread, SIGKILL);
        LOCK_SYSTEM;
        RemoveEntryList(&(q->links));
        free(q);
        UNLOCK_SYSTEM;
        return(TRUE);
    }
    return(FALSE);
}



void  
SystemDebugBreakpoint(void)
/**

Routine description:

    A hardcoded breakpoint.

Parameters:

    void

Return value:

    void

**/
{
#if !defined(_ANDROID)
	__asm__("int3\n");
#endif
}

uint64_t FASTCALL 
SystemReadTimeStampCounter(void)
/**

Routine description:

    Read the processor's timestamp counter.

Parameters:

    void

Return value:

    uint64_t FASTCALL

**/
{
#if !defined(_ANDROID)
    __asm__("rdtsc;"
        :
        :
        : "%eax", "%edx");
#endif
}


double INLINE
SystemTimeStamp(void)
/**

Routine description:

    Return a timestamp to the caller.

Parameters:

    void

Return value:

    double

**/
{
    struct timeval tv;

    //
    // Number of seconds and microseconds since the Epoch.
    //
    if (0 != gettimeofday(&tv, NULL))
    {
        UtilPanic(UNEXPECTED_SYSTEM_CALL_FAILURE,
                  NULL, "gettimeofday", (void *)errno, NULL,
                  __FILE__, __LINE__);
    }
    return((double)tv.tv_sec + (double)tv.tv_usec * 1.0e-6);
}


FLAG
SystemDoesFileExist(int8_t *szFilename)
/**

Routine description:

    Determine if a file exists (non-atomically).

Parameters:

    int8_t *szFilename

Return value:

    FLAG

**/
{
    struct stat sb;
    int32_t i = stat(szFilename, &sb);

    if ((i == -1) && (errno == ENOENT)) return(FALSE);
    return(TRUE);
}


static FLAG 
_SystemIsRoot(void)
/**

Routine description:

    This code is called by SystemDependentInitialization.  Its job is to
    determine whether this process is running with administrative powers.

Parameters:

    void

Return value:

    static FLAG

**/
{
    if (0 == geteuid())
    {
        return(TRUE);
    }
    return(FALSE);
}


void  
SystemDeferExecution(uint32_t dwMs)
/**

Routine description:

    Sleep for a few ms.

Parameters:

    uint32_t dwMs

Return value:

    void 

**/
{
    struct timeval t = { 0, dwMs * 1000 };
    if (select(0, (fd_set *)NULL, (fd_set *)NULL, (fd_set *)NULL, &t) < 0) 
    {
        if (usleep(dwMs * 1000) < 0) 
        {
            Bug("SystemDeferExecution: I have insomnia...\n");
        }
    }
}


static void *
_SystemCallMmap(uint32_t dwSizeBytes)
/**

Routine description:

    Wrapper around mmap.

Parameters:

    uint32_t dwSizeBytes

Return value:

    static void

**/
{
    void *pMem = mmap(0,
                      (size_t)dwSizeBytes,
                      PROT_READ | PROT_WRITE,
                      MAP_ANON,
                      -1,
                      0);
    if (MAP_FAILED == pMem)
    {
        UtilPanic(UNEXPECTED_SYSTEM_CALL_FAILURE,
                  NULL, "mmap", (void *)errno, (void *)dwSizeBytes,
                  __FILE__, __LINE__);
    }
    (void)madvise(pMem, dwSizeBytes, MADV_RANDOM | MADV_WILLNEED);
    return(pMem);
}


static void 
_SystemCallMunmap(void *pMem)
/**

Routine description:

    Wrapper around munmap

Parameters:

    void *pMem

Return value:

    static void

**/
{
    if (0 != munmap(pMem, (size_t)-1))
    {
        UtilPanic(UNEXPECTED_SYSTEM_CALL_FAILURE,
                  NULL, "munmap", (void *)errno, pMem,
                  __FILE__, __LINE__);
    }
}


static void *_SystemCallMalloc(uint32_t dwSizeBytes)
/**

Routine description:

    Wrapper around malloc.

Parameters:

    uint32_t dwSizeBytes

Return value:

    static void

**/
{
    void *p = malloc(dwSizeBytes);
    if (NULL == p)
    {
        UtilPanic(UNEXPECTED_SYSTEM_CALL_FAILURE,
                  NULL, "malloc", (void *)errno, (void *)dwSizeBytes,
                  __FILE__, __LINE__);
    }
    memset(p, 0, dwSizeBytes);
    return(p);
}


static void 
_SystemCallFree(void *p)
/**

Routine description:

    Wrapper around free.

Parameters:

    void *p

Return value:

    static void

**/
{
    free(p);
}

int8_t *
SystemStrDup(int8_t *p)
/**

Routine description:

    Implements strdup functionality.

Parameters:

    int8_t *p

Return value:

    int8_t

**/
{
    int8_t *q = SystemAllocateMemory(strlen(p) + 1);
    strcpy(q, p);
    return(q);
}

void *
SystemAllocateMemory(uint32_t dwSizeBytes)
/**

Routine description:
 
    Allocate some memory.

Parameters:

    uint32_t dwSizeBytes

Return value:

    void

**/
{
    uint32_t *p;

    if (0) // (dwSizeBytes > SYS_MAX_HEAP_ALLOC_SIZE_BYTES)
    {

        #ifdef __LP64__
        p = _SystemCallMmap(dwSizeBytes + sizeof(uint64_t));
#else
        p = _SystemCallMmap(dwSizeBytes + sizeof(uint32_t));
#endif
        if (NULL == p)
        {
            UtilPanic(UNEXPECTED_SYSTEM_CALL_FAILURE,
                      NULL, "mmap", (void *)errno, NULL,
                      __FILE__, __LINE__);
        }
        *p = MMAP;
        p += 1;
    }
    else
    {
        #ifdef __LP64__
		p = _SystemCallMalloc(dwSizeBytes + sizeof(uint64_t));
       #else
        p = _SystemCallMalloc(dwSizeBytes + sizeof(uint32_t));
#endif
        if (NULL == p)
        {
            UtilPanic(UNEXPECTED_SYSTEM_CALL_FAILURE,
                      NULL, "malloc", (void *)errno, NULL,
                      __FILE__, __LINE__);
        }
        *p = HEAP;
        p += 1;
    }
#ifdef DEBUG
    MarkAllocHashEntry(p, dwSizeBytes);
#endif
    ASSERT(NULL != p);
    return(p);
}

void 
SystemFreeMemory(void *pMemory)
/**

Routine description:

    Free an allocation previously allocated by SystemAllocMemory.

Parameters:

    void *pMemory

Return value:

    void

**/
{
    #ifdef __LP64__
	uint64_t *p = pMemory;
    #else
    uint32_t *p = pMemory;
    #endif
    
    p -= 1;
#ifdef DEBUG
    ReleaseAllocHashEntry(pMemory);
#endif
    if (*p == HEAP)
    {
        _SystemCallFree(p);
    }
    else if (*p == MMAP)
    {
        _SystemCallMunmap(p);
    }
#ifdef DEBUG
    else
    {
        UtilPanic(SHOULD_NOT_GET_HERE,
                  NULL, NULL, NULL, NULL,
                  __FILE__, __LINE__);
    }
#endif
}


void *
SystemAllocateLockedMemory(uint32_t dwSizeBytes)
/**

Routine description:

    Allocate non-pagable memory (if possible).

Parameters:

    uint32_t dwSizeBytes

Return value:

    void

**/
{
    void *p = SystemAllocateMemory(dwSizeBytes);
    if (NULL != p)
    {
        if (0 != mlock(p, dwSizeBytes))
        {
            //
            // Note: don't panic the program over this... do dump a 
            // warning though as it should not happen.
            //
            Bug("WARNING: mlock for buffer at 0x%p (%u bytes) failed, "
                "errno=%d\n\n", p, dwSizeBytes, errno);
        }
    }
    return(p);
}


FLAG 
SystemMakeMemoryReadOnly(void *pMemory, 
                         uint32_t dwSizeBytes)
/**

Routine description:

    Make some memory region read-only (if possible).

Parameters:

    void *pMemory,
    uint32_t dwSizeBytes

Return value:

    FLAG

**/
{
    if (0 != mprotect(pMemory, dwSizeBytes, PROT_READ))
    {
        UtilPanic(UNEXPECTED_SYSTEM_CALL_FAILURE,
                  NULL, "mprotect", (void *)errno, (void *)PROT_READ,
                  __FILE__, __LINE__);
    }
    return(TRUE);
}



FLAG 
SystemMakeMemoryReadWrite(void *pMemory, uint32_t dwSizeBytes)
/**

Routine description:

    Make a region of memory read/write.

Parameters:

    void *pMemory,
    uint32_t dwSizeBytes

Return value:

    FLAG

**/
{
    if (0 != mprotect(pMemory, dwSizeBytes, PROT_READ | PROT_WRITE))
    {
        UtilPanic(UNEXPECTED_SYSTEM_CALL_FAILURE,
                  NULL, 
                  "mprotect", 
                  (void *)errno,
                  (void *)(PROT_READ | PROT_WRITE),
                  __FILE__, __LINE__);
    }
    return(TRUE);
}



FLAG 
SystemCopyFile(int8_t *szSource, int8_t *szDest)
/**

Routine description:

    Make a copy of a file.

Parameters:

    int8_t *szSource,
    int8_t *szDest

Return value:

    FLAG

**/
{
    int8_t *buf = NULL;
    int8_t *x = "/bin/cp -f ";
    uint32_t uLen = strlen(szSource) + strlen(szDest) + strlen(x) + 2;

    buf = SystemAllocateMemory(uLen);
    snprintf(buf, uLen, 
             "%s%s %s", x, szSource, szDest);
    if (0 == system(buf))
    {
        SystemFreeMemory(buf);
        return(TRUE);
    }
    SystemFreeMemory(buf);
    return(FALSE);
}


FLAG 
SystemDeleteFile(int8_t *szFile)
/**

Routine description:

    Delete a file.

Parameters:

    int8_t *szFile

Return value:

    FLAG

**/
{
    if (0 == unlink(szFile))
    {
        return(TRUE);
    }
    return(FALSE);
}


#define MAX_LOCKS (8)
typedef struct _UNIX_LOCK_ENTRY
{
    pthread_mutex_t lock;
    FLAG fInUse;
} UNIX_LOCK_ENTRY;
UNIX_LOCK_ENTRY g_rgLockTable[MAX_LOCKS];

uint32_t
SystemCreateLock(void) 
{
    uint32_t u;

    LOCK_SYSTEM;
    for (u = 0; u < MAX_LOCKS; u++)
    {
        if (FALSE == g_rgLockTable[u].fInUse)
        {
            pthread_mutex_init(&(g_rgLockTable[u].lock), NULL);
            g_rgLockTable[u].fInUse = TRUE;
            goto end;
        }
    }
    u = (uint32_t)-1;                            // no free slot

 end:
    UNLOCK_SYSTEM;
    return(u);
}

FLAG
SystemDeleteLock(uint32_t u)
{
    LOCK_SYSTEM;
    if ((u < MAX_LOCKS) && (g_rgLockTable[u].fInUse == TRUE))
    {
        pthread_mutex_destroy(&(g_rgLockTable[u].lock));
        g_rgLockTable[u].fInUse = FALSE;
        UNLOCK_SYSTEM;
        return(TRUE);
    }
    UNLOCK_SYSTEM;
    return(FALSE);
}

FLAG
SystemBlockingWaitForLock(uint32_t u) 
{
    pthread_mutex_t *p = NULL;
    LOCK_SYSTEM;
    if ((u < MAX_LOCKS) && (g_rgLockTable[u].fInUse == TRUE)) 
    {
        p = &(g_rgLockTable[u].lock);
    }
    UNLOCK_SYSTEM;
    if (p != NULL)
    {
        pthread_mutex_lock(p);
        return(TRUE);
    }
    return(FALSE);
}


FLAG
SystemReleaseLock(uint32_t u) 
{
    pthread_mutex_t *p = NULL;
    LOCK_SYSTEM;
    if ((u < MAX_LOCKS) && (g_rgLockTable[u].fInUse == TRUE)) 
    {
        p = &(g_rgLockTable[u].lock);
    }
    UNLOCK_SYSTEM;
    if (p != NULL) 
    {
        pthread_mutex_unlock(p);
        return(TRUE);
    }
    return(FALSE);
}


#define MAX_SEM (8)
int32_t g_rgSemaphores[MAX_SEM];
#ifdef DEBUG
uint32_t g_rgSemValues[MAX_SEM];
#endif

uint32_t
SystemCreateSemaphore(uint32_t uValue) 
{
    uint32_t u;
    union semun {
            int32_t val;
            struct semid_ds *buf;
            unsigned short  *array;
    } value;

    if (uValue > MAX_USHORT) return((uint32_t)-1);
    LOCK_SYSTEM;
    for (u = 0; u < MAX_SEM; u++) 
    {
        if (g_rgSemaphores[u] < 0) 
        {
            g_rgSemaphores[u] = semget(IPC_PRIVATE, 1, IPC_CREAT | 0600);
            if (g_rgSemaphores[u] < 0) 
            {
                u = (uint32_t)-1;
                goto end;
            }
            value.val = uValue;
            if (semctl(g_rgSemaphores[u], 0, SETVAL, value) < 0) 
            {
                (void)SystemDeleteSemaphore(u);
                u = (uint32_t)-1;
                goto end;
            }
#ifdef DEBUG
            g_rgSemValues[u] = uValue;
            Trace("Semaphore #%u created, initial value is %u\n", 
                  u, uValue);
#endif
            goto end;
        }
    }
    u = (uint32_t)-1;
 end:
    UNLOCK_SYSTEM;
    return(u);
}

FLAG 
SystemDeleteSemaphore(uint32_t u) {
    union semun {
            int32_t val;
            struct semid_ds *buf;
            unsigned short  *array;
    } value;

    LOCK_SYSTEM;
    if ((u < MAX_SEM) && (g_rgSemaphores[u] >= 0)) 
    {
        if (semctl(g_rgSemaphores[u], 0, IPC_RMID, value) < 0) 
        {
            return(FALSE);
        }
        g_rgSemaphores[u] = -1;
    }
    UNLOCK_SYSTEM;
    return(TRUE);
}

void
SystemReleaseSemaphoreResource(uint32_t u) 
{
    struct sembuf operation;

    LOCK_SYSTEM;
    if ((u < MAX_SEM) && (g_rgSemaphores[u] >= 0)) 
    {
        operation.sem_num = u;
        operation.sem_op = +1;
        operation.sem_flg = 0;
        if (semop(g_rgSemaphores[u], &operation, 1) < 0) 
        {
            UtilPanic(UNEXPECTED_SYSTEM_CALL_FAILURE,
                      NULL, "semop", (void *)errno, (void *)0,
                      __FILE__, __LINE__);
        }
#ifdef DEBUG
        g_rgSemValues[u] += 1;
        Trace("Semaphore #%u value incremented to %u\n", u, 
              g_rgSemValues[u]);
#endif
    }
    UNLOCK_SYSTEM;
}

void
SystemObtainSemaphoreResource(uint32_t u) 
{
    struct sembuf operation;

    LOCK_SYSTEM;
    if ((u < MAX_SEM) && (g_rgSemaphores[u] >= 0))
    {
        operation.sem_num = u;
        operation.sem_op = -1;
        operation.sem_flg = 0;
        UNLOCK_SYSTEM;
        if (semop(g_rgSemaphores[u], &operation, 1) < 0) 
        {
            UtilPanic(UNEXPECTED_SYSTEM_CALL_FAILURE,
                      NULL, "semop", (void *)errno, (void *)1,
                      __FILE__, __LINE__);
        }
#ifdef DEBUG
        if (g_rgSemValues[u] > 0) 
        {
            g_rgSemValues[u] -= 1;
            Trace("Semaphore #%u value decremented to %u\n", u, 
                  g_rgSemValues[u]);
        }
#endif
        return;
    }
    UNLOCK_SYSTEM;
}


FLAG
SystemDependentInitialization(void)
/**

Routine description:

    This code is called to initialize operating system dependent code.
    It is unsynchronized.  It should only ever be called once.  It
    should definitely never be called by two threads at once!

Parameters:

    void

Return value:

    FLAG : TRUE on success, FALSE otherwise

**/
{
    uint32_t u;
#ifdef DEBUG
    static FLAG fInit = FALSE;

    if (TRUE == fInit)
    {
        Bug("SystemDependentInitialization code called more than once!\n");
        ASSERT(FALSE);
        return(FALSE);
    }
    fInit = TRUE;

    memset(g_AllocHash, 0, sizeof(g_AllocHash));
#endif
    for (u = 0; u < MAX_SEM; u++) 
    {
        g_rgSemaphores[u] = -1;
    }
    InitializeListHead(&g_SystemThreadList);
    pthread_mutex_init(&g_SystemLock, NULL);
    return(TRUE);
}    


#else

#include "chess.h"
#define _WIN32_WINNT 0x403

#pragma warning(disable:4142) // benign redefinition of type
#include <windows.h>
#pragma warning(default:4142)

#define SYS_MAX_HEAP_ALLOC_SIZE_BYTES 0xfff

//
// Global system lock
//
CRITICAL_SECTION g_SystemLock;
#define LOCK_SYSTEM   EnterCriticalSection(&g_SystemLock)
#define UNLOCK_SYSTEM LeaveCriticalSection(&g_SystemLock)

//
// Thread table
//
typedef struct _SYSTEM_THREAD_TABLE_ENTRY
{
    DLIST_ENTRY links;
    uint32_t uWrapperHandle;
    uint32_t uTid;
    uint32_t uThreadParam;
    THREAD_ENTRYPOINT *pEntry;
    uint32_t uExitCode;
    HANDLE hThread;
}
SYSTEM_THREAD_TABLE_ENTRY;
DLIST_ENTRY g_SystemThreadList;

//
// Global buffer to hold a bunch of system dependent numbers / settings
// 
typedef struct _SYS_INFORMATION_BUFFER
{
    FLAG fPopulated;
    SYSTEM_INFO si;
    OSVERSIONINFOEX vi;
    MEMORYSTATUS ms;
}
SYS_INFORMATION_BUFFER;
SYS_INFORMATION_BUFFER g_SystemInfoBuffer;

//
// The frequency and type of our TimeStamp timer
//
double g_dTimeStampFrequency = 0.0;
uint32_t g_uTimeStampType = 0;
uint32_t g_uMaxHeapAllocationSize = 0;

//
// A history of allocations
// 
#ifdef DEBUG
uint32_t g_uTotalAlloced = 0;

#define ALLOC_HASH_SIZE 0x4000

typedef struct _ALLOC_RECORD
{
    void *p;
    uint32_t uSize;
} ALLOC_RECORD;
ALLOC_RECORD g_AllocHash[ALLOC_HASH_SIZE];

#define PTR_TO_ALLOC_HASH(x) ((((ULONG_PTR)(x)) >> 3) & (ALLOC_HASH_SIZE - 1))

uint32_t
GetHeapMemoryUsage(void)
{
    Trace("There are now %u bytes outstanding.\n", g_uTotalAlloced);
    return(g_uTotalAlloced);
}

static void 
ReleaseAllocHashEntry(void *p)
{
    uint32_t u = PTR_TO_ALLOC_HASH(p);
    uint32_t v = u;

    LOCK_SYSTEM;
    while(g_AllocHash[u].p != p)
    {
        u++;
        u &= (ALLOC_HASH_SIZE - 1);
        if (v == u) return;
    }
    g_AllocHash[u].p = NULL;
    g_uTotalAlloced -= g_AllocHash[u].uSize;
    UNLOCK_SYSTEM;

//    Trace("Freed %u bytes at 0x%p.\n", g_AllocHash[u].uSize, p);
}

static void
MarkAllocHashEntry(void *p, uint32_t uSize)
{
    uint32_t u = PTR_TO_ALLOC_HASH(p);

    if (uSize > (1024 * 1024 * 30)) return;

    LOCK_SYSTEM;
    while(g_AllocHash[u].p != NULL)
    {
        u++;
        u &= (ALLOC_HASH_SIZE - 1);
    }
    g_AllocHash[u].p = p;
    g_AllocHash[u].uSize = uSize;
    g_uTotalAlloced += uSize;
    UNLOCK_SYSTEM;

//    Trace("Alloc %u bytes at 0x%p.\n", g_AllocHash[u].uSize, p);
}
#endif


int8_t *
SystemStrDup(int8_t *p)
/**

Routine description:

    Local implementation of strdup

Parameters:

    int8_t *p

Return value:

    int8_t

**/
{
    int8_t *q = SystemAllocateMemory((uint32_t)strlen(p) + 1);

    ASSERT(strlen(p) < (size_t)MAX_ULONG);
    strcpy(q, p);
    return(q);
}


int8_t *
SystemGetDateString(void)
/**

Routine description:

    Get the current date as a string.

Parameters:

    void

Return value:

    int8_t

**/
{
    static int8_t buf[32];
    SYSTEMTIME st;
    
    GetSystemTime(&st);

    snprintf(buf, ARRAY_LENGTH(buf), 
             "%u.%02u.%02u", st.wYear, st.wMonth, st.wDay);
    return(buf);
}


int8_t *
SystemGetTimeString(void)
/**

Routine description:

    Get the current time as a string.

Parameters:

    void

Return value:

    int8_t

**/
{
    static int8_t buf[32];
    SYSTEMTIME st;
    
    GetSystemTime(&st);

    snprintf(buf, ARRAY_LENGTH(buf),
             "%u:%02u:%02u", st.wHour, st.wMinute, st.wSecond);
    return(buf);
}


void 
SystemDeferExecution(uint32_t uMs)
/**

Routine description:

    Sleep for some number of milliseconds.

Parameters:

    uint32_t uMs

Return value:

    void

**/
{
    Sleep(uMs);
}

static SYSTEM_THREAD_TABLE_ENTRY *
_SystemGetThreadTableEntry(uint32_t uThreadHandle)
/**

Routine description:

    Support code for the thread functionality wrapper.  Given a thread
    "handle" return its thread table struct.

Parameters:

    uint32_t uThreadHandle

Return value:

    SYSTEM_THREAD_TABLE_ENTRY *

**/
{
    DLIST_ENTRY *p;
    SYSTEM_THREAD_TABLE_ENTRY *q;

    LOCK_SYSTEM;
    p = g_SystemThreadList.pFlink;
    while(p != &g_SystemThreadList)
    {
        q = CONTAINING_STRUCT(p, SYSTEM_THREAD_TABLE_ENTRY, links);
        if (q->uWrapperHandle == uThreadHandle)
        {
            goto end;
        }
        p = p->pFlink;
    }
    q = NULL;
 end:
    UNLOCK_SYSTEM;
    return(q);
}

int32_t WINAPI 
SystemThreadEntryPoint(VOID *pParam)
/**

Routine description:

    The real entry point of all new threads.  It calls into the user
    specified entry point.

Parameters:

    VOID *pParam

Return value:

    int32_t WINAPI

**/
{
    SYSTEM_THREAD_TABLE_ENTRY *p = (SYSTEM_THREAD_TABLE_ENTRY *)pParam;
    uint32_t uParam;
    int32_t i;
    
    LOCK_SYSTEM;
    UNLOCK_SYSTEM;

    uParam = p->uThreadParam;
    i = (int32_t)(*(p->pEntry))(uParam);
    return(i);
}

FLAG 
SystemCreateThread(THREAD_ENTRYPOINT *pEntry, uint32_t uParam, uint32_t *puHandle)
/**

Routine description:

    Wrapper function to expose create thread functionality.

Parameters:

    THREAD_ENTRYPOINT *pEntry,
    uint32_t uParam,
    uint32_t *puHandle

Return value:

    FLAG

**/
{
    SYSTEM_THREAD_TABLE_ENTRY *p;

    p = HeapAlloc(GetProcessHeap(),
                  HEAP_ZERO_MEMORY,
                  sizeof(SYSTEM_THREAD_TABLE_ENTRY));
    if (NULL == p)
    {
        *puHandle = 0;
        return(FALSE);
    }

    LOCK_SYSTEM;
    p->pEntry = pEntry;
    p->uThreadParam = uParam;
    p->hThread = CreateThread(NULL,
                              0x6000,
                              SystemThreadEntryPoint,
                              (VOID *)p,
                              0,
                              &(p->uTid));
    if (NULL == p->hThread)
    {
        (void)HeapFree(GetProcessHeap(), 0, p);
        return(FALSE);
    }
    *puHandle = p->uWrapperHandle = p->uTid;
    InsertHeadList(&g_SystemThreadList, &(p->links));
    UNLOCK_SYSTEM;
    return(TRUE);
}


FLAG 
SystemWaitForThreadToExit(uint32_t uThreadHandle)
/**

Routine description:

    Wait for the specified thread to exit.

Parameters:

    uint32_t uThreadHandle

Return value:

    FLAG

**/
{
    SYSTEM_THREAD_TABLE_ENTRY *q = _SystemGetThreadTableEntry(uThreadHandle);
    
    if (WAIT_OBJECT_0 != WaitForSingleObject(q->hThread, INFINITE))
    {
        ASSERT(FALSE);
        return(FALSE);
    }
    return(TRUE);
}


FLAG 
SystemGetThreadExitCode(uint32_t uThreadHandle, uint32_t *puCode)
/**

Routine description:

    Get the specified thread's exit value.

Parameters:

    uint32_t uThreadHandle,
    uint32_t *puCode

Return value:

    FLAG

**/
{
    SYSTEM_THREAD_TABLE_ENTRY *q = _SystemGetThreadTableEntry(uThreadHandle);
    uint32_t uCode;
    
    if (FALSE == GetExitCodeThread(q->hThread, &uCode))
    {
        return(FALSE);
    }
    if (STILL_ACTIVE == uCode)
    {
        return(FALSE);
    }
    *puCode = uCode;
    return(TRUE);
}


FLAG 
SystemDeleteThread(uint32_t uThreadHandle)
/**

Routine description:

    Terminate a thread.

Parameters:

    uint32_t uThreadHandle

Return value:

    FLAG

**/
{
    SYSTEM_THREAD_TABLE_ENTRY *q = _SystemGetThreadTableEntry(uThreadHandle);
    uint32_t u;
    
    if (NULL != q)
    {
        if (FALSE == SystemGetThreadExitCode(uThreadHandle, &u))
        {
            SystemWaitForThreadToExit(uThreadHandle);
        }
        CloseHandle(q->hThread);
        LOCK_SYSTEM;
        RemoveEntryList(&(q->links));
        HeapFree(GetProcessHeap(), 0, q);
        UNLOCK_SYSTEM;
        return(TRUE);
    }
    return(FALSE);
}


static void
_SystemPopulateSystemInformationBuffer(void)
/**

Routine description:
    
    Populates global structure SYS_INFORMATION_BUFFER.  This code should
    be called only once from SystemDependentInitialization.  It should
    definitely never be called by more than one thread at a time.

Parameters:

    void

Return value:

    void

**/
{
#ifdef DEBUG
    static FLAG fInit = FALSE;

    if (TRUE == fInit)
    {
        Bug("_SystemPopulateSystemInformationBuffer: This code should "
            "not be called more than once!\n");
        ASSERT(FALSE);
    }
    fInit = TRUE;
#endif

    //
    // Make some calls to populate the global buffer
    //
    GetSystemInfo(&(g_SystemInfoBuffer.si));
    
    // 
    // Note: This API is not present on Win9x
    //
    g_SystemInfoBuffer.vi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
    if (FALSE == GetVersionEx((OSVERSIONINFO *)&(g_SystemInfoBuffer.vi)))
    {
        Bug("_SystemPopulateSystemInformationBuffer: GetVersionEx "
            "call failed, error=%u.\n", GetLastError());
        ASSERT(FALSE);
    }

    GlobalMemoryStatus(&(g_SystemInfoBuffer.ms));
    
    g_SystemInfoBuffer.fPopulated = TRUE;
}


static FLAG 
_SystemEnablePrivilege(TCHAR *szPrivilegeName,
                       FLAG fEnable)
/**

Routine description:

    This code is called by SystemDependentInitialization to add some
    privileges to our process' security token.

Parameters:

    TCHAR *szPrivilegeName,
    FLAG fEnable

Return value:

    FLAG

**/
{
    FLAG fMyRetVal = FALSE;                   // our return value
    HANDLE hToken = INVALID_HANDLE_VALUE;     // handle to process token
    PTOKEN_PRIVILEGES pNewPrivileges;         // privs to enable/disable
    LUID Luid;                                // priv as a LUID
    uint8_t buf[sizeof(TOKEN_PRIVILEGES)+
            ((1 - ANYSIZE_ARRAY) * sizeof(LUID_AND_ATTRIBUTES))];

    pNewPrivileges = (PTOKEN_PRIVILEGES)buf;
    
    //
    // Try to open the thread's security token first.
    // 
    if (FALSE == OpenThreadToken(GetCurrentThread(),
                                 TOKEN_ADJUST_PRIVILEGES,
                                 FALSE,
                                 &hToken))
    {
        //
        // Well that failed.  If it is because there is no thread token then
        // try to open the process' security token.
        //
        if (ERROR_NO_TOKEN == GetLastError())
        {
            if (FALSE == OpenProcessToken(GetCurrentProcess(),
                                          TOKEN_ADJUST_PRIVILEGES,
                                          &hToken))
            {
                //
                // Can't open the process token either.  Fail.
                // 
                goto end;
            }
        }
        else
        { 
            //
            // Failed to open the thread token for some other reason.  Fail.
            // 
            goto end;
        }
    }
    
    //
    // Convert priv name to an LUID.
    //
    if (FALSE == LookupPrivilegeValue(NULL,
                                      szPrivilegeName,
                                      &Luid))
    {
        goto end;
    }
    
    //
    // Construct new data struct to enable / disable the privi.
    //
    pNewPrivileges->PrivilegeCount = 1;
    pNewPrivileges->Privileges[0].Luid = Luid;
    pNewPrivileges->Privileges[0].Attributes = 
        fEnable ? SE_PRIVILEGE_ENABLED : 0;

    //
    // Adjust the privileges on the token.
    //
    fMyRetVal = AdjustTokenPrivileges(hToken,
                                      FALSE,
                                      pNewPrivileges,
                                      0,
                                      NULL,
                                      NULL);
    if (TRUE == fMyRetVal)
    {
        //
        // This boneheaded API returns TRUE and then expects the
        // caller to test GetLastError to see if it really worked.
        // 
        if (ERROR_SUCCESS != GetLastError())
        {
            fMyRetVal = FALSE;
        }
    }
    
 end:
    if (INVALID_HANDLE_VALUE != hToken)
    {
        CloseHandle(hToken);
    }
    return(fMyRetVal);
}


void  
SystemDebugBreakpoint(void)
/**

Routine description:

    A hardcoded breakpoint.

Parameters:

    void

Return value:

    void

**/
{
    DebugBreak();
}

#define SYS_HIGH_PERFORMANCE_TIMER (1)
#define SYS_GET_TICK_COUNT         (2)
#define SYS_MS_PER_SECOND          (1000)

double
SystemTimeStamp(void)
/**

Routine description:

    Return a timestamp to the caller.  The precision of this timestamp
    is based on whether the machine supports a high resolution timer.
    If not, this code rolls back to use GetTickCount().

    This code relies on some initialization code in SystemDependent-
    Initialization.  Once that code has executed, this code is thread
    safe.

Parameters:

    void

Return value:

    double

**/
{
    double dRetVal;
    uint64_t u64;

    switch(g_uTimeStampType)
    {
        case SYS_HIGH_PERFORMANCE_TIMER:
            QueryPerformanceCounter((LARGE_INTEGER *)&u64);
            dRetVal = u64 * g_dTimeStampFrequency;
            break;
        case SYS_GET_TICK_COUNT:
            dRetVal = GetTickCount() / SYS_MS_PER_SECOND;
            break;
        default:
            dRetVal = 0.0;
            Bug("SystemTimeStamp: You have to call SystemDependent"
                "Initialization before you can call this code.\n");
            ASSERT(FALSE);
            break;
    }
    return(dRetVal);
}


FLAG 
SystemDoesFileExist(int8_t *szFilename)
/**

Routine description:

    Determine if a file exists (non-atomically).

Parameters:

    int8_t *szFilename

Return value:

    FLAG

**/
{
    if (INVALID_FILE_ATTRIBUTES == GetFileAttributesA(szFilename))
    {
        if (GetLastError() == ERROR_FILE_NOT_FOUND)
        {
            return(FALSE);
        }
    }
    return(TRUE);
}


static FLAG 
_SystemIsAdministrator(void)
/**

Routine description:

    This code is called by SystemDependentInitialization.  Its job is to
    determine whether this process is running with administrative powers.

Parameters:

    void

Return value:

    static FLAG

**/
{
    BOOL fIsMember = FALSE;                   // our return value
    SID *pAdminSid = NULL;                    // ptr to admin sid
    SID_IDENTIFIER_AUTHORITY NtAuthority =    // used to create admin sid
        SECURITY_NT_AUTHORITY;
    
    //
    // Allocate and initialize a SID for the admin group.
    // 
    if (FALSE == AllocateAndInitializeSid(&NtAuthority,
                                          2,
                                          SECURITY_BUILTIN_DOMAIN_RID,
                                          DOMAIN_ALIAS_RID_ADMINS,
                                          0, 0, 0, 0, 0, 0,
                                          &pAdminSid)) 
    {
        goto end;
    }
    
    //
    // SID allocation succeeded, check this thread or process token
    // for membership in the admin group.
    // 
    if (FALSE == CheckTokenMembership(NULL,
                                      pAdminSid,
                                      &fIsMember))
    {
        fIsMember = FALSE;
    }
    
 end:
    if (NULL != pAdminSid)
    {
        FreeSid(pAdminSid);
    }
    
    return(fIsMember);
}


static void *
_SystemCallVirtualAlloc(uint32_t dwSizeBytes)
/**

Routine description:

    Wrapper around VirtualAlloc

Parameters:

    uint32_t dwSizeBytes

Return value:

    static void

**/
{
    SIZE_T size = dwSizeBytes;
    void *pMem = NULL;
    
    pMem = VirtualAlloc(NULL,
                        size,
                        MEM_RESERVE | MEM_COMMIT, 
                        PAGE_READWRITE);
    if (NULL == pMem)
    {
        UtilPanic(UNEXPECTED_SYSTEM_CALL_FAILURE,
                  NULL, "VirtualAlloc", GetLastError(), size,
                  __FILE__, __LINE__);
    }
    return(pMem);
}

static void 
_SystemCallVirtualFree(void *pMem)
/**

Routine description:

    Wrapper around VirtualFree

Parameters:

    void *pMem

Return value:

    static void

**/
{
    BOOL fRetVal = FALSE;

    if (NULL != pMem)
    {
        fRetVal = VirtualFree(pMem, 0, MEM_RELEASE);
        if (FALSE == fRetVal)
        {
            UtilPanic(UNEXPECTED_SYSTEM_CALL_FAILURE,
                      NULL, "VirtualFree", GetLastError(), pMem,
                      __FILE__, __LINE__);
        }
    }
}

static void *
_SystemCallHeapAlloc(uint32_t dwSizeBytes)
/**

Routine description:

    Wrapper around HeapAlloc.  Returns a zeroed buffer from the heap.
    While on the subject of heaps, let me take a moment to extol the
    value of the Win32 Application Verifier / PageHeap code.  This is
    code that sanity checks the actions of an app.  It is especially
    good at finding heap misuse (double frees, reuse after free,
    buffer overruns, etc...)  It also catches things like
    CRITICAL_SECTION abuse and HANDLE abuse.  To enable it, use
    appverif.exe.

Parameters:

    uint32_t dwSizeBytes

Return value:

    static void

**/
{
    void *p = HeapAlloc(GetProcessHeap(),
                        HEAP_ZERO_MEMORY,
                        dwSizeBytes);
    if (NULL == p)
    {
        UtilPanic(UNEXPECTED_SYSTEM_CALL_FAILURE,
                  NULL, "HeapAlloc", GetLastError(), dwSizeBytes,
                  __FILE__, __LINE__);
    }
    return(p);
}

static void 
_SystemCallHeapFree(void *pMem)
/**

Routine description:

    Wrapper around HeapFree

Parameters:

    void *pMem

Return value:

    static void

**/
{
    if (FALSE == HeapFree(GetProcessHeap(),
                          0,
                          pMem))
    {
        UtilPanic(UNEXPECTED_SYSTEM_CALL_FAILURE,
                  NULL, "HeapFree", GetLastError(), pMem,
                  __FILE__, __LINE__);
    }
}

void 
SystemFreeMemory(void *pMem)
/**

Routine description:

    Free memory

Parameters:

    void *pMem

Return value:

    static void

**/
{
    _SystemCallHeapFree(pMem);
#ifdef DEBUG
    ReleaseAllocHashEntry(pMem);
#endif
}

void *
SystemAllocateMemory(uint32_t dwSizeBytes)
/**

Routine description:

    Allocate some memory, either from the heap or from VirtualAlloc
    depending on the size of the request.  The buffer returned by
    this call is guaranteed to be zeroed out.  Buffers allocated by
    this call should be freed via a call to SystemFreeMemory when
    they are no longer needed.

Parameters:

    uint32_t dwSizeBytes : size of the buffer needed in bytes

Return value:

    void

**/
{
    uint8_t *p;

    p = _SystemCallHeapAlloc(dwSizeBytes);
#ifdef DEBUG
    MarkAllocHashEntry(p, dwSizeBytes);
#endif
    return(p);
}


void *
SystemAllocateLockedMemory(uint32_t dwSizeBytes)
/**

Routine description:

    Allocate a buffer that is locked in memory (i.e. non-pagable).

Parameters:

    uint32_t dwSizeBytes

Return value:

    void

**/
{
    void *p = _SystemCallVirtualAlloc(dwSizeBytes);

    if (NULL != p)
    {
        if (FALSE == VirtualLock(p, dwSizeBytes))
        {
            UtilPanic(UNEXPECTED_SYSTEM_CALL_FAILURE,
                      NULL, "VirtualLock", GetLastError(), p,
                      __FILE__, __LINE__);
        }
    }
#ifdef DEBUG
    MarkAllocHashEntry(p, dwSizeBytes);
#endif
    return(p);
}


FLAG 
SystemMakeMemoryReadOnly(void *pMemory, 
                         uint32_t dwSizeBytes)
/**

Routine description:

Parameters:

    void *pMemory,
    uint32_t dwSizeBytes

Return value:

    FLAG

**/
{
    BOOL fRetVal = FALSE;
    uint32_t dwJunk;
    
    fRetVal = VirtualProtect(pMemory, 
                             dwSizeBytes, 
                             PAGE_READONLY, 
                             &dwJunk);
    if (FALSE == fRetVal)
    {
        UtilPanic(UNEXPECTED_SYSTEM_CALL_FAILURE,
                  NULL, "VirtualProtect", GetLastError(), pMemory,
                  __FILE__, __LINE__);
    }
    return((FLAG)fRetVal);
}

FLAG 
SystemMakeMemoryNoAccess(void *pMemory, 
                         uint32_t dwSizeBytes)
/**

Routine description:

Parameters:

    void *pMemory,
    uint32_t dwSizeBytes

Return value:

    FLAG

**/
{
    BOOL fRetVal = FALSE;
    uint32_t dwJunk;
    
    fRetVal = VirtualProtect(pMemory, 
                             dwSizeBytes, 
                             PAGE_NOACCESS,
                             &dwJunk);
    if (FALSE == fRetVal)
    {
        UtilPanic(UNEXPECTED_SYSTEM_CALL_FAILURE,
                  NULL, "VirtualProtect", GetLastError(), pMemory,
                  __FILE__, __LINE__);
    }
    return((FLAG)fRetVal);
}


FLAG 
SystemCopyFile(int8_t *szSource, int8_t *szDest)
/**

Routine description:

Parameters:

    int8_t *szSource,
    int8_t *szDest

Return value:

    FLAG

**/
{
    return(CopyFileA(szSource, szDest, TRUE));
}


FLAG 
SystemDeleteFile(int8_t *szFile)
/**

Routine description:

Parameters:

    int8_t *szFile

Return value:

    FLAG

**/
{
    return(DeleteFileA(szFile));
}


FLAG 
SystemMakeMemoryReadWrite(void *pMemory, uint32_t dwSizeBytes)
/**

Routine description:

Parameters:

    void *pMemory,
    uint32_t dwSizeBytes

Return value:

    FLAG

**/
{
    BOOL fRetVal = FALSE;
    uint32_t dwJunk;
    
    fRetVal = VirtualProtect(pMemory, 
                             dwSizeBytes, 
                             PAGE_READWRITE, 
                             &dwJunk);
    if (FALSE == fRetVal)
    {
        Bug("SystemMakeMemoryReadWrite: VirtualProtect for buffer at %p "
            "(size %u) failed, error=%u.\n", pMemory, dwSizeBytes,
            GetLastError());
    }
    return(fRetVal);
}


uint64_t FASTCALL 
SystemReadTimeStampCounter(void)
/**

Routine description:

    Read the processor's timestamp counter.

Parameters:

    void

Return value:

    uint64_t FASTCALL

**/
{
#if !defined(_ANDROID)
	__asm__("rdtsc;"
        :
        :
        : "%eax", "%edx");
#endif
}



#define MAX_LOCKS (8)
typedef struct _WIN32_LOCK_ENTRY
{
    CRITICAL_SECTION lock;
    FLAG fInUse;
} WIN32_LOCK_ENTRY;
WIN32_LOCK_ENTRY g_rgLockTable[MAX_LOCKS];

uint32_t
SystemCreateLock(void) 
{
    uint32_t u;

    LOCK_SYSTEM;
    for (u = 0; u < MAX_LOCKS; u++)
    {
        if (FALSE == g_rgLockTable[u].fInUse)
        {
            InitializeCriticalSection(&(g_rgLockTable[u].lock));
            g_rgLockTable[u].fInUse = TRUE;
            goto end;
        }
    }
    u = (uint32_t)-1;                            // no free slot

 end:
    UNLOCK_SYSTEM;
    return(u);
}

FLAG
SystemDeleteLock(uint32_t u)
{
    LOCK_SYSTEM;
    if ((u < MAX_LOCKS) && (g_rgLockTable[u].fInUse == TRUE)) 
    {
        DeleteCriticalSection(&(g_rgLockTable[u].lock));
        g_rgLockTable[u].fInUse = FALSE;
        UNLOCK_SYSTEM;
        return(TRUE);
    }
    UNLOCK_SYSTEM;
    return(FALSE);
}

FLAG
SystemBlockingWaitForLock(uint32_t u) 
{
    CRITICAL_SECTION *p = NULL;
    LOCK_SYSTEM;
    if ((u < MAX_LOCKS) && (g_rgLockTable[u].fInUse == TRUE)) 
    {
        p = &(g_rgLockTable[u].lock);
    }
    UNLOCK_SYSTEM;
    if (p != NULL) 
    {
        EnterCriticalSection(p);
        Trace("lock at %p is obtained.\n", p);
        return(TRUE);
    }
    return(FALSE);
}


FLAG
SystemReleaseLock(uint32_t u) 
{
    CRITICAL_SECTION *p = NULL;
    LOCK_SYSTEM;
    if ((u < MAX_LOCKS) && (g_rgLockTable[u].fInUse == TRUE)) 
    {
        p = &(g_rgLockTable[u].lock);
    }
    UNLOCK_SYSTEM;
    if (p != NULL) 
    {
        LeaveCriticalSection(p);
        Trace("lock at %p is released.\n", p);
        return(TRUE);
    }
    return(FALSE);
}


#define MAX_SEMS (8)
HANDLE g_rgSemHandles[MAX_SEMS];

uint32_t 
SystemCreateSemaphore(uint32_t uValue) 
{
    uint32_t u;
    LOCK_SYSTEM;
    for (u = 0; u < MAX_SEMS; u++) 
    {
        if (g_rgSemHandles[u] == (HANDLE)NULL) 
        {
            g_rgSemHandles[u] = CreateSemaphore(NULL, uValue, uValue, NULL);
            if (NULL == g_rgSemHandles[u]) 
            {
                u = (uint32_t)-1;
            }
            goto end;
        }
    }
    u = (uint32_t)-1;

 end:
    UNLOCK_SYSTEM;
    return(u);
}

FLAG
SystemDeleteSemaphore(uint32_t u) 
{
    LOCK_SYSTEM;
    if ((u < MAX_SEMS) && (g_rgSemHandles[u] != (HANDLE)NULL)) 
    {
        (void)CloseHandle(g_rgSemHandles[u]);
        g_rgSemHandles[u] = (HANDLE)NULL;
        UNLOCK_SYSTEM;
        return(TRUE);
    }
    UNLOCK_SYSTEM;
    return(FALSE);
}

void
SystemReleaseSemaphoreResource(uint32_t u) 
{
    LOCK_SYSTEM;
    if ((u < MAX_SEMS) && (g_rgSemHandles[u] != (HANDLE)NULL)) 
    {
        ReleaseSemaphore(g_rgSemHandles[u], +1, NULL);
    }
    UNLOCK_SYSTEM;
}

void
SystemObtainSemaphoreResource(uint32_t u) 
{
    HANDLE h;
    LOCK_SYSTEM;
    if ((u < MAX_SEMS) && (g_rgSemHandles[u] != (HANDLE)NULL)) 
    {
        h = g_rgSemHandles[u];
        UNLOCK_SYSTEM;
        (void)WaitForSingleObject(h, INFINITE);
        return;
    }
    UNLOCK_SYSTEM;
}

      

FLAG 
SystemDependentInitialization(void)
/**

Routine description:

    This code is called to initialize operating system dependent code.
    It is unsynchronized.  It should only ever be called once.  It
    should definitely never be called by two threads at once!

Parameters:

    void

Return value:

    FLAG : TRUE on success, FALSE otherwise

**/
{
    uint64_t u64;
#ifdef DEBUG
    static FLAG fInit = FALSE;

    if (TRUE == fInit)
    {
        Bug("SystemDependentInitialization code called more than once!\n");
        ASSERT(FALSE);
        return(FALSE);
    }
    fInit = TRUE;
#endif

    //
    // Initialize the system dependent lock
    //
#if defined(__try)
    __try

    {
#endif
        InitializeCriticalSection(&g_SystemLock);
 
#if defined(__try)
	}
 
	__except(EXCEPTION_EXECUTE_HANDLER)
    {
        Bug("SystemDependentInitialization: InitializeCriticalSection "
            "raised an exception, bailing out.\n");
        return(FALSE);
    }
#endif

    //
    // Clear a pool of handles
    //
    memset(g_rgLockTable, 0, sizeof(g_rgLockTable));
    memset(g_rgSemHandles, 0, sizeof(g_rgSemHandles));

    //
    // Populate global system info buffer
    //
    _SystemPopulateSystemInformationBuffer();

    //
    // Enable some privileges in this process' security token
    //
    if (FALSE == _SystemEnablePrivilege(SE_INC_BASE_PRIORITY_NAME, TRUE))
    {
        Trace("NOTICE: Can't enable SE_INC_BASE_PRIORITY_NAME privilege, "
              "error=%u\n", GetLastError());
    }
    if (FALSE == _SystemEnablePrivilege(SE_LOCK_MEMORY_NAME, TRUE))
    {
#if !defined(_WIN32) && !defined(_WIN64)
		Trace("NOTICE: Can't enable SE_LOCK_MEMORY_NAME privilege, "
              "error=%u\n", GetLastError());
#endif
    }
    
    //
    // Don't let the system hibernate/suspend while the chess program
    // is running.
    // 
    (void)SetThreadExecutionState(ES_CONTINUOUS | ES_SYSTEM_REQUIRED);

    //
    // Determine the precision of our SystemTimeStamp() code
    //
    ASSERT(g_uTimeStampType == 0);
    if (FALSE != QueryPerformanceFrequency((LARGE_INTEGER *)&u64))
    {
        g_dTimeStampFrequency = 1.0 / u64;
        g_uTimeStampType = SYS_HIGH_PERFORMANCE_TIMER;
    }
    else
    {
        g_dTimeStampFrequency = 1.0 / 1000.0;
        g_uTimeStampType = SYS_GET_TICK_COUNT;
    }
    ASSERT(g_dTimeStampFrequency > 0.0);

    g_uMaxHeapAllocationSize = 0x1000;
    InitializeListHead(&g_SystemThreadList);

    return(TRUE);
}    


#endif
